//
// Module: RegionExtent
// Author: Hassan Siddiqui
//
// Description: 
//
// Class to determine the DSS region extents in
// X/Y and DETX/Y
//
// Modifications: See ChangeLog
// 
// $Id: RegionExtent.cc,v 1.2 2011/01/07 16:43:36 rsaxton Exp $

/* SAS */
#include "errstr.h"
#include "Tracer.h"
#include "Xmm.h"

/* Local */
#include "RegionExtent.h"
#include "PositionUtils.h"
#include "PosWCS.h"

/* Construct object given explicit extent */
RegionExtent::RegionExtent( double xmin,
			    double xmax,
			    double ymin,
			    double ymax ):
  _isValid(true),
  _xmin(xmin),
  _xmax(xmax),
  _ymin(ymin),
  _ymax(ymax)
{
  PROC;
  ENDPROC;
}


/* Construct object based given a DataSubspace object */
RegionExtent::RegionExtent( DataSubspace * dss, 
     PositionUtils * putils, const string & datamode, 
     const string & dateobs, const PosWCS * const poswcs ):
  _isValid(false)
{

  PROC;
  
  /* Get extents in DETX/DETY space */
  double lower_detx;
  double upper_detx;
  RangeBound lowerB_detx;
  RangeBound upperB_detx;
  Range<double> extent_detx = dss->extent("DETX");
  extent_detx.values(lower_detx, 
		     upper_detx, 
		     lowerB_detx, 
		     upperB_detx);

  errstr 
    << "DETX extents : "
    << " lower_detx=" << lower_detx
    << " lowerB_detx=" << (unsigned) lowerB_detx
    << " upper_detx=" << upper_detx
    << " upperB_detx=" << (unsigned) upperB_detx
    << Sas::message(Msg::App, Msg::Noisy);

  
  double lower_dety;
  double upper_dety;
  RangeBound lowerB_dety;
  RangeBound upperB_dety;
  Range<double> extent_dety = dss->extent("DETY");
  extent_dety.values(lower_dety, 
		     upper_dety, 
		     lowerB_dety, 
		     upperB_dety);

  errstr 
    << "DETY extents : "
    << " lower_dety=" << lower_dety
    << " lowerB_dety=" << (unsigned) lowerB_dety
    << " upper_dety=" << upper_dety
    << " upperB_dety=" << (unsigned) upperB_dety
    << Sas::message(Msg::App, Msg::Noisy);

  /* Get extents in X/Y space */
  double lower_posx;
  double upper_posx;
  RangeBound lowerB_posx;
  RangeBound upperB_posx;
  Range<double> extent_posx = dss->extent("X");
  extent_posx.values(lower_posx, 
		     upper_posx, 
		     lowerB_posx, 
		     upperB_posx);
  errstr 
    << "POSX extents : "
    << " lower_posx=" << lower_posx
    << " lowerB_posx=" << (unsigned) lowerB_posx
    << " upper_posx=" << upper_posx
    << " upperB_posx=" << (unsigned) upperB_posx
    << Sas::message(Msg::App, Msg::Noisy);

  
  double lower_posy;
  double upper_posy;
  RangeBound lowerB_posy;
  RangeBound upperB_posy;
  Range<double> extent_posy = dss->extent("Y");
  extent_posy.values(lower_posy, 
		  upper_posy, 
		  lowerB_posy, 
		  upperB_posy);

  errstr 
    << "POSY extents : "
    << " lower_posy=" << lower_posy
    << " lowerB_posy=" << (unsigned) lowerB_posy
    << " upper_posy=" << upper_posy
    << " upperB_posy=" << (unsigned) upperB_posy
    << Sas::message(Msg::App, Msg::Noisy);

      
  /* Now check whether regions in either coordinate
     system is bounded. */
  bool detUnbounded = ( lowerB_detx == Unbounded ||
			upperB_detx == Unbounded ||
			lowerB_dety == Unbounded ||
			upperB_dety == Unbounded );
  
  bool posUnbounded = ( lowerB_posx == Unbounded ||
			upperB_posx == Unbounded ||
			lowerB_posy == Unbounded ||
			upperB_posy == Unbounded );
  
  if ( detUnbounded && posUnbounded )
    {
      errstr 
	<< "DSS regions defined in X/Y and DETX/DETY space are unbounded. The ARF can not be calculated"
	<< Sas::error("RegionInvalid");

      ALTRETURN;
    }

  _xmin = lower_detx;
  _xmax = upper_detx;
  _ymin = lower_dety;
  _ymax = upper_dety;
      
  if (! posUnbounded )
    {
      /* Get pos coords mapped onto DET */
      
      SourcePosition slowerleft = putils->sourcePosition( "POS",
                                                     datamode,
                                                     dateobs,
                                                     lower_posx,
                                                     lower_posy,
                                                     poswcs);
      SourcePosition slowerright = putils->sourcePosition( "POS",
                                                     datamode,
                                                     dateobs,
                                                     upper_posx,
                                                     lower_posy,
                                                     poswcs);
      SourcePosition supperleft = putils->sourcePosition( "POS",
                                                     datamode,
                                                     dateobs,
                                                     lower_posx,
                                                     upper_posy,
                                                     poswcs);
      SourcePosition supperright = putils->sourcePosition( "POS",
                                                     datamode,
                                                     dateobs,
                                                     upper_posx,
                                                     upper_posy,
                                                     poswcs);

     /* Get min/max values */
      vector<double> xlist;
      xlist.push_back(slowerleft .xdet());
      xlist.push_back(slowerright.xdet());
      xlist.push_back(supperleft .xdet());
      xlist.push_back(supperright.xdet());

      vector<double> ylist;
      ylist.push_back(slowerleft .ydet());
      ylist.push_back(slowerright.ydet());
      ylist.push_back(supperleft .ydet());
      ylist.push_back(supperright.ydet());

      double lower_posx_det = xlist[0];
      double upper_posx_det = xlist[0];
      double lower_posy_det = ylist[0];
      double upper_posy_det = ylist[0];

      for (unsigned int i = 1; i < xlist.size(); ++i)
        {
          if (xlist[i] < lower_posx_det)
            lower_posx_det = xlist[i];

          if (xlist[i] > upper_posx_det)
            upper_posx_det = xlist[i];

          if (ylist[i] < lower_posy_det)
            lower_posy_det = ylist[i];

          if (ylist[i] > upper_posy_det)
            upper_posy_det = ylist[i];
        }

      errstr
        << "RegionExtent: from POS region spec, found extent in DET coords to be " << endl
        << " lower_posx_det " << lower_posx_det
        << " upper_posx_det " << upper_posx_det
        << " lower_posy_det " << lower_posy_det
        << " upper_posy_det " << upper_posy_det
        << Sas::message(Msg::App, Msg::Verbose);

      /* [HS] STOP */

      _xmin = lower_posx_det;
      _ymin = lower_posy_det;
      _xmax = upper_posx_det;
      _ymax = upper_posy_det;

      if (! detUnbounded)
	{
	  /* Get the larger of the bounds from
	     the DET and POS regions (NB used to be the
	     smaller of the bounds - which was a bug */

	  //if (lower_detx > lower_posx_det)
	  if (lower_detx < lower_posx_det)
	    _xmin = lower_detx;
	  
	  //if (upper_detx < upper_posx_det)
	  if (upper_detx > upper_posx_det)
	    _xmax = upper_detx;
	  
	  //if (lower_dety > lower_posy_det)
	  if (lower_dety < lower_posy_det)
	    _ymin = lower_dety;
	  
	  //if (upper_dety < upper_posy_det)
	  if (upper_dety > upper_posy_det)
	    _ymax = upper_dety;
	}
    }

  errstr
    << "RegionExtent: FINAL extent computed is " << endl
    << " xmin " << _xmin
    << " xmax " << _xmax
    << " ymin " << _ymin
    << " ymax " << _ymax
    << Sas::message(Msg::App, Msg::Verbose);

  /* [HS] STOP expand code */

  _isValid = true;

  ENDPROC;
}

/* Check if object is constructed correctly */
bool RegionExtent::isValid() const
{
  PROC;
  RETURN _isValid;
}

/* Report if extent is larger than specified bounds */
void RegionExtent::checkBounds( double boundsxmin,
				double bounds_xmax,
				double bounds_ymin,
				double bounds_ymax) const
{
  PROC;

  errstr 
    << "DSS x-axis extent ["
    << _xmin
    << ","
    << _xmax
    << "]" << endl
    << "Detmap x-axis bounds [" 
    << boundsxmin
    << ","
    << bounds_xmax
    << "]."
    << Sas::message(Msg::App, Msg::Noisy);
  
  if ( _xmin  < boundsxmin ||
       _xmax > bounds_xmax )
    {
      errstr 
	<< "The DSS x-axis extent ["
	<< _xmin
	<< ","
	<< _xmax
	<< "] exceeds the detmap x-axis bounds [" 
	<< boundsxmin
	<< ","
	<< bounds_xmax
	<< "]. This may produce an inappropriate ARF"
	<< Sas::warning("detmapXBoundsExceeded");
    }
  
  errstr 
    << "DSS y-axis extent ["
    << _ymin
    << ","
    << _ymax
    << "]" << endl
    << "Detmap y-axis bounds [" 
    << bounds_ymin
    << ","
    << bounds_ymax
    << "]."
    << Sas::message(Msg::App, Msg::Noisy);
  
  if ( _ymin < bounds_ymin ||
       _ymax > bounds_ymax )
    {
      errstr 
	<< "The DSS y-axis extent ["
	<< _ymin
	<< ","
	<< _ymax
	<< "] exceeds the detmap y-axis bounds [" 
	<< bounds_ymin
	<< ","
	<< bounds_ymax
	<< "]. This may produce an inappropriate ARF"
	<< Sas::warning("detmapYBoundsExceeded");
    }

  ENDPROC;
}

/* Accessors */

double RegionExtent::xmin() const
{
  PROC; 
  RETURN _xmin;
}

double RegionExtent::xmax() const
{
  PROC; 
  RETURN _xmax;
}

double RegionExtent::ymin() const
{
  PROC; 
  RETURN _ymin;
}

double RegionExtent::ymax() const
{
  PROC; 
  RETURN _ymax;
}
